####Clustering for detecting proteins of concordant change
####Chen et al. (2017), JASA.

library(mvtnorm)
##library(grpreg)
##install.packages("gglasso")
library(gglasso)
library(ggplot2)
library(MASS)

multiplot <-
  function(...,
           plotlist = NULL,
           file,
           cols = 1,
           layout = NULL) {
    library(grid)
    
    # Make a list from the ... arguments and plotlist
    plots <- c(list(...), plotlist)
    
    numPlots = length(plots)
    
    # If layout is NULL, then use 'cols' to determine layout
    if (is.null(layout)) {
      # Make the panel
      # ncol: Number of columns of plots
      # nrow: Number of rows needed, calculated from # of cols
      layout <- matrix(seq(1, cols * ceiling(numPlots / cols)),
                       ncol = cols,
                       nrow = ceiling(numPlots / cols))
    }
    
    if (numPlots == 1) {
      print(plots[[1]])
      
    } else {
      # Set up the page
      grid.newpage()
      pushViewport(viewport(layout = grid.layout(nrow(layout), ncol(layout))))
      
      # Make each plot, in the correct location
      for (i in 1:numPlots) {
        # Get the i,j matrix positions of the regions that contain this subplot
        matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))
        
        print(plots[[i]],
              vp = viewport(
                layout.pos.row = matchidx$row,
                layout.pos.col = matchidx$col
              ))
      }
    }
  }


CorrAR <- function(p, rho) {
  Sigma <- matrix(nrow = p, ncol = p, NA)
  for (i in 1:p) {
    for (j in 1:p) {
      Sigma[i, j] <- rho ^ (abs(i - j))
    }
  }
  Sigma
}

CorrCS <- function(p, rho) {
  Sigma <- matrix(nrow = p, ncol = p, rho)
  diag(Sigma) <- 1
  Sigma
}


##Soft thresholding for rows
softrowTH <- function(C, Lam, tol = 1e-4) {
  l2row <-
    sqrt(rowSums(C ^ 2)) # apply(C, 1, function(x) sqrt(sum(x^2)))
  under <- l2row - Lam < tol
  C[under,] <- 0
  C[!under,] <- C[!under,] * (1 - Lam / l2row)[!under]
  C
}

##Hard thresholding for rows
hardrowTH <- function(C, lambda, eta = 0) {
  l2row <- apply(C, 1, function(x)
    sqrt(sum(x ^ 2)))
  C <- C / (1 + eta)
  C[l2row <= lambda, ] <- 0
  list(C = C, l2 = l2row)
}

##Control function
mixture.control <- function(tol = 1e-6,
                            maxit = 200,
                            plot = TRUE,
                            trace = TRUE,
                            plotname = "Rplots.pdf",
                            sign.cons = FALSE,
                            mu.cons = FALSE,
                            sigma.type = c("diag", "corr")[2],
                            iter.inner = 1,
                            update.gamma.first = FALSE,
                            warm.start = FALSE) {
  if (!is.numeric(tol) || tol <= 0)
    stop("value of 'tol' must be > 0")
  if (!is.numeric(maxit) || maxit <= 0)
    stop("maximum number of iterations must be > 0")
  if (!is.numeric(iter.inner) || iter.inner <= 0)
    stop("number of inner iterations must be > 0")
  list(
    tol = tol,
    maxit = maxit,
    plot = plot,
    trace = trace,
    plotname = plotname,
    sign.cons = sign.cons,
    mu.cons = mu.cons,
    sigma.type = sigma.type,
    iter.inner = iter.inner,
    update.gamma.first = update.gamma.first,
    warm.start = warm.start
  )
}

## Objective function
mixture.obj <-
  function(Y,
           lambda,
           Theta = list(),
           method = c("glasso", "ghard")[2]) {
    dimY <- dim(Y)
    P <- dimY[1]
    M <- dimY[2]
    S <- dimY[3]
    K <- 3
    Mu <- Theta$Mu
    Sigma <- Theta$Sigma
    Alpha <- Theta$Alpha
    Gamma <- Theta$Gamma
    Pi <- Theta$Pi
    
    dnorm_k <- matrix(nrow = P, ncol = K)
    ##For the 1th component
    dnorm_s <- matrix(nrow = P, ncol = S)
    for (s in 1:S) {
      dnorm_s[, s] <-
        apply(Y[, , s] - Gamma, 1, function(a)
          mvtnorm::dmvnorm(a, Alpha[s] + Mu[, 1], Sigma[, , 1]))
    }
    dnorm_k[, 1] <- apply(dnorm_s, 1, prod) * Pi[K]
    for (k in 2:K) {
      for (s in 1:S) {
        dnorm_s[, s] <-
          apply(Y[, , s], 1, function(a)
            mvtnorm::dmvnorm(a, Alpha[s] + Mu[, k], Sigma[, , k]))
      }
      dnorm_k[, k] <- apply(dnorm_s, 1, prod) * Pi[k]
    }
    loglike <- sum(log(apply(dnorm_k, 1, sum)))
    if (method == "glasso") {
      pen <- lambda * sum(apply(Gamma, 1, function(a)
        sqrt(sum(a ^ 2))))
    }
    return(loglike - pen)
  }



## main function for clustering
mixture.penalized <-
  function(Y,
           lambda,
           ini = list(),
           method = c("glasso", "ghard")[2],
           control = list()) {
    control <- do.call("mixture.control", control)
    
    dimY <- dim(Y)
    P <- dimY[1]
    M <- dimY[2]
    S <- dimY[3]
    
    ##All parameters: Pi, Mu, Sigma, Alpha and Gamma
    Theta0 <-
      c(
        ini$Pi,
        as.vector(ini$Mu),
        as.vector(ini$Sigma),
        as.vector(ini$Alpha),
        as.vector(ini$Gamma)
      )
    Theta <- Theta0
    ##initial values
    Mu <- ini$Mu
    Pi <- ini$Pi
    Sigma <- ini$Sigma
    Alpha <- ini$Alpha
    Gamma <- ini$Gamma
    tol <- control$tol
    maxit <- control$maxit
    
    
    diff <- vector()
    obj <- vector()
    loglike <- vector()
    iter <- 1
    diff[iter] <- tol * 2
    
    ##Get the index sets for observed data
    for (s in 1:S) {
      assign(paste("id", s, sep = ""), which(!is.na(rowSums(Y[, , s]))))
    }
    ###Number for each p
    srepnum <- apply(apply(Y, c(1, 3), sum), 1, function(a)
      sum(!is.na(a)))
    
    ## Used when mean constrain is adopted.
    Mu_obj <- function(mu, k) {
      b <- 0
      for (s in 1:S) {
        ###Get the ids
        ids <- get(paste("id", s, sep = ""))
        a <- t(apply(Y[ids, , s] - Alpha[s], 1, function(a)
          a - mu))
        b <- b + sum((a %*% Sigma2inv * Pik[ids, k]) * a)
      }
      b / 2
    }
    Mu.deri <- function(mu, k) {
      b <- rep(0, M)
      for (s in 1:S) {
        ###Get the ids
        ids <- get(paste("id", s, sep = ""))
        a <- t(apply(Y[ids, , s] - Alpha[s], 1, function(a)
          a - mu))
        b <- b + colSums((a %*% Sigma2inv * Pik[ids, k]))
      }
      - b
    }
    
    Ytemp <- Y
    
    ## Used when shared correlation constrain is adopted
    if (control$sigma.type == "corr") {
      Ak.obj <- function(a, Sigma1inv, Yk, divisor) {
        Ak <- diag(a, nrow = M, ncol = M)
        - divisor * log(prod(a)) + 1 / 2 * sum(diag(Ak %*% Sigma1inv %*% Ak %*%
                                                      Yk))
      }
      Ak.derv <- function(a, Sigma1inv, Yk, divisor) {
        Ak <- diag(a, nrow = M, ncol = M)
        bb <- Sigma1inv %*% Ak %*% Yk
        - divisor / a + diag(bb + t(bb)) / 2
      }
      ##
      A <- array(dim = c(M, M, K - 1), 0)
      for (k in 2:K) {
        A[, , k - 1] <-
          diag((diag(Sigma[, , k]) / diag(Sigma[, , 1])) ^ (1 / 2), nrow = M, ncol =
                 M)
      }
    }
    
    
    
    
    while (iter < maxit & diff[iter] > tol) {
      Theta0 <- Theta
      
      ############Update Pik################################
      dnorm_k <- matrix(nrow = P, ncol = K)
      ##For the 1th component
      dnorm_s <- matrix(nrow = P, ncol = S)
      for (s in 1:S) {
        dnorm_s[, s] <-
          mvtnorm::dmvnorm(Y[, , s] - Gamma, Alpha[s] + Mu[, 1], Sigma[, , 1], log = TRUE)
      }
      dnorm_k[, 1] <- exp(rowSums(dnorm_s, na.rm = TRUE)) * Pi[1]
      for (k in 2:K) {
        for (s in 1:S) {
          dnorm_s[, s] <-
            mvtnorm::dmvnorm(Y[, , s], Alpha[s] + Mu[, k], Sigma[, , k], log = TRUE)
        }
        dnorm_k[, k] <- exp(rowSums(dnorm_s, na.rm = TRUE)) * Pi[k]
      }
      Pik <- dnorm_k / rowSums(dnorm_k)
      ##May produce NA in extreme cases
      na.id <- which(apply(Pik, 1, function(a)
        sum(is.na(a))) != 0)
      if (length(na.id) != 0) {
        Pik[na.id, ] <- matrix(nrow = length(na.id),
                               ncol = K,
                               Pi,
                               byrow = TRUE)
      }
      ##compute current objective
      loglike[iter] <- sum(log(rowSums(dnorm_k)))
      if (method == "glasso") {
        ##This is based on gglasso package
        ##Their obj is -loglike/n + lam*glasso
        pen <- lambda * sum(sqrt(rowSums(Gamma ^ 2))) * sqrt(M) * M
        obj[iter] <- loglike[iter] - pen
      }
      if (method == "ghard") {
        Noutlier <- sum(rowSums(abs(Gamma)) != 0)
        pen <- lambda ^ 2 / 2 * Noutlier
        obj[iter] <- loglike[iter] - pen
      }
      
      
      #################Update pi################################
      Pi <- colSums(Pik) / P
      
      
      
      
      if (control$update.gamma.first == TRUE) {
        #########Update Gamma###################
        eigenSig1 <- eigen(Sigma[, , 1])
        Sig1sqrtinv <-
          eigenSig1$vectors %*% diag(1 / sqrt(eigenSig1$values)) %*% t(eigenSig1$vectors)
        
        for (s in 1:S) {
          Ytemp[, , s] <- Y[, , s] - Alpha[s]
        }
        GrpY0 <- t(rowMeans(Ytemp, dims = 2, na.rm = TRUE))
        GrpY <- Sig1sqrtinv %*% GrpY0
        GrpX <- Sig1sqrtinv
        if (method == "glasso") {
          ###The objective is RSS/(2*n) + lambda * penalty;
          lambda_g <- lambda / Pik[, 1] / srepnum
          ##Indices of zero gamma vectors
          zero_index <-
            which(sqrt(apply(GrpX %*% GrpY, 2, function(a)
              sum(a ^ 2))) <= M * lambda_g * sqrt(M))
          if (length(zero_index) > 0) {
            Gamma[zero_index, ] <- 0
            nonzero_index <- (1:P)[-zero_index]
          } else{
            nonzero_index <- 1:P
          }
          for (i in nonzero_index) {
            fiti <- gglasso(
              GrpX,
              y = GrpY[, i],
              group = rep(1, M),
              nlambda = 1,
              lambda = lambda_g[i],
              intercept = FALSE
            )
            Gamma[i, ] <- fiti$beta
          }
          
          
        }
        
        if (method == "ghard") {
          l2rowsq <- colSums(GrpY ^ 2)
          l2lambda <- lambda ^ 2 / Pik[, 1] / srepnum
          Gamma[l2rowsq <= l2lambda, ] <- 0
          l2index <- (l2rowsq > l2lambda)
          Gamma[l2index, ]  <- t(GrpY0[, l2index])
        }
        control$update.gamma.first <- FALSE
      }
      
      
      
      
      
      
      
      #For inner loop
      for (niter in 1:control$iter.inner) {
        ################Update Alpha##############################
        for (s in 1:S) {
          ###Get the ids
          ids <- get(paste("id", s, sep = ""))
          
          sum_ik <- matrix(nrow = length(ids), ncol = K, NA)
          sinv_k <- rep(1, K)
          for (k in 2:K) {
            Sigmakinv <- ginv(Sigma[, , k])
            Yidss <- Y[ids, , s]
            Yids2 <- matrix(nrow = M, ncol = length(ids))
            for (j in 1:M) {
              Yids2[j, ] <- Yidss[, j] - Mu[j, k]
            }
            sum_ik[, k] <- colSums(Sigmakinv %*% Yids2)
            sinv_k[k] <- sum(Sigmakinv)
          }
          
          ##For the 1th component
          Sigma1inv <- ginv(Sigma[, , 1])
          sum_ik[, 1] <- colSums(Sigma1inv %*% t(Y[ids, , s]))
          sinv_k[1] <- sum(Sigma1inv)
          Alpha[s] <-
            sum(Pik[ids, ] * sum_ik, na.rm = TRUE) / sum(apply(Pik[ids, ], 1, function(a)
              a * sinv_k))
        }
        
        
        ################Update Mu###################################
        Sigma2inv <- ginv(Sigma[, , 2])
        Sigma3inv <- ginv(Sigma[, , 3])
        
        if (control$mu.cons == TRUE) {
          sum_low <- matrix(nrow = M, ncol = S)
          sum_high <- matrix(nrow = M, ncol = S)
          divisor <- matrix(nrow = M, ncol = M, 0)
          for (s in 1:S) {
            ###Get the ids
            ids <- get(paste("id", s, sep = ""))
            k <- 2
            sum_low[, s] <-
              colSums((Y[ids, , s] - Alpha[s]) %*% Sigma2inv * Pik[ids, k])
            k <- 3
            sum_high[, s] <-
              colSums((Y[ids, , s] - Alpha[s]) %*% Sigma3inv * Pik[ids, k])
            divisor <-
              divisor + sum(Pik[ids, 2]) * Sigma2inv + sum(Pik[ids, 3]) * Sigma3inv
          }
          Mu[, 2] <- ginv(divisor) %*% rowSums(sum_low - sum_high)
          Mu[, 3] <- -Mu[, 2]
          
        } else{
          sum_low <- matrix(nrow = M, ncol = S)
          sum_high <- matrix(nrow = M, ncol = S)
          divisor_low <- matrix(nrow = M, ncol = M, 0)
          divisor_high <- matrix(nrow = M, ncol = M, 0)
          for (s in 1:S) {
            ###Get the ids
            ids <- get(paste("id", s, sep = ""))
            k <- 2
            sum_low[, s] <-
              colSums((Y[ids, , s] - Alpha[s]) %*% Sigma2inv * Pik[ids, k])
            
            k <- 3
            sum_high[, s] <-
              colSums((Y[ids, , s] - Alpha[s]) %*% Sigma3inv * Pik[ids, k])
            
            divisor_low <- divisor_low + sum(Pik[ids, 2]) * Sigma2inv
            divisor_high <- divisor_high + sum(Pik[ids, 3]) * Sigma3inv
          }
          
          Mu2temp <- ginv(divisor_low) %*% rowSums(sum_low)
          Mu3temp <- ginv(divisor_high) %*% rowSums(sum_high)
          
          if (control$sign.cons == TRUE) {
            if (sum(Mu2temp < 0) != M) {
              Mu2temp <-
                constrOptim(
                  Mu[, 2] - 1e-4,
                  f = function(a)
                    Mu_obj(a, 2),
                  ##grad = NULL,
                  grad = function(a)
                    Mu.deri(a, 2),
                  ui = -diag(M),
                  ci = rep(0, M)
                )$par
            }
            if (sum(Mu3temp > 0) != M) {
              Mu3temp <-
                constrOptim(
                  Mu[, 3] + 1e-4,
                  f = function(a)
                    Mu_obj(a, 3),
                  ##grad = NULL,
                  grad = function(a)
                    Mu.deri(a, 3),
                  ui = diag(M),
                  ci = rep(0, M)
                )$par
            }
          }
          Mu[, 2] <- Mu2temp
          Mu[, 3] <- Mu3temp
          
          
        }
        
        
        #################Update Sigma################################
        if (control$sigma.type == "diag") {
          ##For updating 2nd and 3rd Sigma, assuming they are diagonal
          for (k in 2:K) {
            sum_s <- matrix(nrow = M, ncol = S)
            divisor <- 0
            for (s in 1:S) {
              ids <- get(paste("id", s, sep = ""))
              
              
              Yidss <- Y[ids, , s]
              Yids2 <- matrix(nrow = M, ncol = length(ids))
              for (j in 1:M) {
                Yids2[j, ] <- (Yidss[, j] - Alpha[s] - Mu[j, k]) ^ 2
              }
              sum_s[, s] <-
                apply(apply(Yids2, 1, function(a)
                  a * Pik[ids, k]), 2, sum)
              divisor <- divisor + sum(Pik[ids, k])
            }
            Sigma[, , k] <- diag(nrow = M, ncol = M) * apply(sum_s, 1, sum) /
              divisor
          }
          
          ###For updating sigma for the reference
          sum_s <- array(dim = c(M, M, S))
          divisor <- 0
          for (s in 1:S) {
            ids <- get(paste("id", s, sep = ""))
            ss <-
              apply(apply(Y[ids, , s] - Gamma[ids, ] - Alpha[s], 1, function(a)
                a - Mu[, 1]), 1, function(a)
                  a * sqrt(Pik[ids, 1]))
            sum_s[, , s] <- crossprod(ss, ss)
            divisor <- divisor + sum(Pik[ids, 1])
          }
          Sigma[, , 1] <- apply(sum_s, c(1, 2), sum) / divisor
        }
        
        ###For assuming the same correlation matrix
        ###In this case Sigma2 and Sigma3 are diagnonal matrices.
        ###For updating sigma for the reference
        if (control$sigma.type == "corr") {
          sum_s <- array(dim = c(M, M, S, K))
          divisor <- 0
          for (s in 1:S) {
            ids <- get(paste("id", s, sep = ""))
            Yidss <- Y[ids, , s]
            #system.time(
            ss <-
              apply(apply(Yidss - Gamma[ids, ] - Alpha[s], 1, function(a)
                a - Mu[, 1]), 1, function(a)
                  a * sqrt(Pik[ids, 1]))
            sum_s[, , s, 1] <- crossprod(ss, ss)
            
            for (k in 2:K) {
              ss <-
                apply(apply(Yidss - Alpha[s], 1, function(a)
                  a - Mu[, k]), 1, function(a)
                    a * sqrt(Pik[ids, k]))
              ss <- ss %*% A[, , k - 1]
              sum_s[, , s, k] <- crossprod(ss, ss)
            }
            divisor <- divisor + length(ids)
          }
          
          Sigma[, , 1] <- apply(sum_s, c(1, 2), sum) / divisor
          
          ###Update Sigma2 and Sigma3
          Sigma1inv <- ginv(Sigma[, , 1])
          for (k in 2:K) {
            sum_s <- array(dim = c(M, M, S))
            divisor <- 0
            for (s in 1:S) {
              ids <- get(paste("id", s, sep = ""))
              Yidss <- Y[ids, , s]
              ss <-
                apply(apply(Yidss - Alpha[s], 1, function(a)
                  a - Mu[, k]), 1, function(a)
                    a * sqrt(Pik[ids, k]))
              sum_s[, , s] <- crossprod(ss, ss)
              divisor <- divisor + sum(Pik[ids, k])
            }
            Yk <- apply(sum_s, c(1, 2), sum)
            
            diag(A[, , k - 1]) <-
              constrOptim(
                diag(A[, , k - 1]),
                f = function(a)
                  Ak.obj(a, Sigma1inv, Yk, divisor),
                #grad = NULL,
                grad = function(a)
                  Ak.derv(a, Sigma1inv, Yk, divisor),
                ui = diag(M),
                ci = rep(0, M)
              )$par
            Ak <- ginv(A[, , k - 1])
            Sigma[, , k] <- Ak %*% Sigma[, , 1] %*% Ak
          }
          
        }
        
        
        
        
        
        #########Update Gamma##############################
        eigenSig1 <- eigen(Sigma[, , 1])
        Sig1sqrtinv <-
          eigenSig1$vectors %*% diag(1 / sqrt(eigenSig1$values)) %*% t(eigenSig1$vectors)
        
        for (s in 1:S) {
          Ytemp[, , s] <- Y[, , s] - Alpha[s]
        }
        GrpY0 <- t(rowMeans(Ytemp, dims = 2, na.rm = TRUE))
        
        
        GrpY <- Sig1sqrtinv %*% GrpY0
        GrpX <- Sig1sqrtinv
        
        if (method == "glasso") {
          ###The objective is RSS/(2*n) + lambda * penalty;
          lambda_g <- lambda / Pik[, 1] / srepnum
          ##Indices of zero gamma vectors
          zero_index <-
            which(sqrt(apply(GrpX %*% GrpY, 2, function(a)
              sum(a ^ 2))) <= M * lambda_g * sqrt(M))
          if (length(zero_index) > 0) {
            Gamma[zero_index, ] <- 0
            nonzero_index <- (1:P)[-zero_index]
          } else{
            nonzero_index <- 1:P
          }
          for (i in nonzero_index) {
            fiti <- gglasso(
              GrpX,
              y = GrpY[, i],
              group = rep(1, M),
              nlambda = 1,
              lambda = lambda_g[i],
              intercept = FALSE
            )
            Gamma[i, ] <- fiti$beta
          }
        }
        
        if (method == "ghard") {
          l2rowsq <- colSums(GrpY ^ 2)
          l2lambda <- lambda ^ 2 / Pik[, 1] / srepnum
          Gamma[l2rowsq <= l2lambda, ] <- 0
          l2index <- (l2rowsq > l2lambda)
          Gamma[l2index, ]  <- t(GrpY0[, l2index])
        }
        
        
        ##For inner loop
      }
      
      
      
      Theta <-
        c(Pi,
          as.vector(Mu),
          as.vector(Sigma),
          as.vector(Alpha),
          as.vector(Gamma))
      iter <- iter + 1
      diff[iter] <- sum((Theta - Theta0) ^ 2) / sum((Theta0) ^ 2)
      
      if (control$trace == TRUE) {
        Noutlier <- sum(rowSums(abs(Gamma)) != 0)
        cat(
          "diff = ",
          round(diff[iter], 4),
          " loglike = ",
          round(loglike[iter - 1], 4),
          " obj = ",
          round(obj[iter - 1], 4),
          " nout = ",
          Noutlier,
          "\n"
        )
      }
      
    }
    
    ###########Graphics#################
    if (control$plot == TRUE) {
      graphics.off()
      pdf(file = control$plotname)
      Y_avg <-
        rbind(Y[id1, , 1] - Alpha[1], Y[id2, , 2] - Alpha[2], Y[id3, , 3] - Alpha[3])
      #classi <- apply(Pik,1,which.max)
      classi <-
        apply(Pik, 1, function(a)
          ifelse(length(which(a > 0.7)) == 1, which(a > 0.7), 5))
      classi[apply(Gamma, 1, function(a)
        sum(a != 0)) != 0] <- 4
      classi <- c(classi[id1], classi[id2], classi[id3])
      
      plots <- list()
      h <- 1
      for (ii in 1:(M - 1)) {
        for (jj in ((ii + 1):M)) {
          dat <- data.frame(x = Y_avg[, ii],
                            y = Y_avg[, jj],
                            z = factor(classi))
          plots[[h]] <- ggplot(dat, aes(
            x = x,
            y = y,
            col = z,
            shape = z
          )) +
            geom_point(alpha = 0.6) +
            theme_bw() +
            theme(legend.position = "none") +
            xlab(paste("VAR ", ii, sep = "")) +
            ylab(paste("VAR ", jj, sep = "")) +
            xlim(range(Y_avg)) +
            ylim(range(Y_avg)) +
            geom_hline(yintercept = 0, colour = "grey") +
            geom_vline(xintercept = 0, colour = "grey") +
            geom_abline(intercept = 0,
                        slope = 1,
                        colour = "grey")
          h <- h + 1
        }
      }
      dev.off()
    }
    
    
    #######Compute obj, loglike and information criteria
    dnorm_k <- matrix(nrow = P, ncol = K)
    ##For the 1th component
    dnorm_s <- matrix(nrow = P, ncol = S)
    for (s in 1:S) {
      dnorm_s[, s] <-
        mvtnorm::dmvnorm(Y[, , s] - Gamma, Alpha[s] + Mu[, 1], Sigma[, , 1], log = TRUE)
    }
    dnorm_k[, 1] <- exp(rowSums(dnorm_s, na.rm = TRUE)) * Pi[1]
    for (k in 2:K) {
      for (s in 1:S) {
        dnorm_s[, s] <-
          mvtnorm::dmvnorm(Y[, , s], Alpha[s] + Mu[, k], Sigma[, , k], log = TRUE)
      }
      dnorm_k[, k] <- exp(rowSums(dnorm_s, na.rm = TRUE)) * Pi[k]
    }
    loglike[iter] <- sum(log(apply(dnorm_k, 1, sum)))
    
    if (method == "glasso") {
      ##This is based on gglasso package
      ##Their obj is -loglike/n + lam*glasso
      pen <-
        lambda * sum(apply(Gamma, 1, function(a)
          sqrt(sum(a ^ 2)))) * sqrt(M) * M
      obj[iter] <- loglike[iter] - pen
    }
    if (method == "ghard") {
      Noutlier <- sum(rowSums(abs(Gamma)) != 0)
      pen <- lambda ^ 2 / 2 * Noutlier
      obj[iter] <- loglike[iter] - pen
    }
    
    ##degrees of freedom of the model:
    Noutlier <- sum(rowSums(abs(Gamma)) != 0)
    if (control$mu.cons == TRUE) {
      df1 <- (S) + (M) + (M * (M - 1) / 2 + M + M) + (K - 1)
    } else{
      df1 <- (S) + (2 * M) + (M * (M - 1) / 2 + M + M) + (K - 1)
    }
    df2 <-
      (Noutlier + sum(sqrt(rowSums(Gamma ^ 2) / colSums(GrpY0 ^ 2))) * (M - 1))
    
    aic <- -2 * loglike[iter] + 2 * df1 + 2 * df2 * P * M / (P * M - df2 -
                                                               1) ##+ 2*risk
    bic <- -2 * loglike[iter] + 2 * (df1 + df2) #+ 1*risk
    
    rval <-
      list(
        Pik = Pik,
        Gamma = Gamma,
        Alpha = Alpha,
        Mu = Mu,
        Sigma = Sigma,
        Pi = Pi,
        diff = diff,
        loglike = loglike,
        obj = obj,
        df = df1 + df2,
        aic = aic,
        bic = bic,
        nout = Noutlier
      )
    class(rval) <- "mixture"
    invisible(rval)
  }



## For fitting the solution path
mixture.penalized.path <-
  function(Y,
           lambda,
           ini = list(),
           method = c("glasso", "ghard")[2],
           control = list()) {
    control <- do.call("mixture.control", control)
    plotind <- control$plot
    plotname <- sapply(strsplit(control$plotname, "\\."), `[[`, 1)
    traceind <- control$trace
    ##When passing to inner, supress trace
    control$trace <- FALSE
    nlambda <- length(lambda)
    
    dimY <- dim(Y)
    P <- dimY[1]
    M <- dimY[2]
    S <- dimY[3]
    
    if (is.null(ini)) {
      #########Using K-means to get initial values################
      ##Adjust for replication effects, and take average
      Alpha_ini <- apply(Y, c(3), function(a)
        mean(a, na.rm = TRUE))
      Y_avg <-
        apply(Y, c(1, 2), function(a)
          mean(a - Alpha_ini, na.rm = TRUE))
      
      Mu_ini <- matrix(nrow = M,
                       ncol = K,
                       c(0, 0, 0, 0, -3, -3, -3, -3, 3, 3, 3, 3))
      inikmeans <- kmeans(Y_avg, centers = t(Mu_ini))
      ini <- list()
      cluster1id <-
        which.min(apply(inikmeans$centers, 1, function(a)
          sum((a - Mu_ini[, 1]) ^ 2)))
      cluster2id <-
        which.min(apply(inikmeans$centers, 1, function(a)
          sum((a - Mu_ini[, 2]) ^ 2)))
      cluster3id <-
        which.min(apply(inikmeans$centers, 1, function(a)
          sum((a - Mu_ini[, 3]) ^ 2)))
      
      ini$Pi <- inikmeans$size[c(cluster1id, cluster2id, cluster3id)] / P
      ini$Mu <- matrix(nrow = M, ncol = K, c(rep(0, M), rep(NA, 2 * M)))
      ini$Mu[, 2] <-  inikmeans$centers[cluster2id, ]
      ini$Mu[, 3] <-  inikmeans$centers[cluster3id, ]
      
      ##library(robust)
      Sigma_ref <- cov(Y_avg[inikmeans$cluster == cluster1id, ])
      Sigma_low <- cov(Y_avg[inikmeans$cluster == cluster2id, ])
      Sigma_high <- cov(Y_avg[inikmeans$cluster == cluster3id, ])
      ini$Sigma <- array(dim = c(M, M, K))
      ini$Sigma[, , 1] <- diag(diag(Sigma_ref))
      ini$Sigma[, , 2] <- diag(diag(Sigma_low))
      ini$Sigma[, , 3] <- diag(diag(Sigma_high))
      ##Replication effects
      ini$Alpha <- Alpha_ini
      ##Mean shift vectors
      ini$Gamma <- matrix(nrow = P, ncol = M, 0)
    }
    
    
    
    Gamma.path <- array(dim = c(nlambda, P, M), NA)
    Alpha.path <- matrix(nrow = nlambda, ncol = S, NA)
    Mu.path <- array(dim = c(nlambda, M, 2), NA)
    Sigma1.path <- array(dim = c(nlambda, M, M), NA)
    Sigma2.path <- matrix(nrow = nlambda, ncol = M, NA)
    Sigma3.path <- matrix(nrow = nlambda, ncol = M, NA)
    Pi.path <- matrix(nrow = nlambda, ncol = K, NA)
    loglike <- vector()
    obj <- vector()
    aic <- vector()
    bic <- vector()
    nout <- vector()
    df <- vector()
    
    for (i in 1:nlambda) {
      control$plotname <- paste(plotname, "_lam", i, ".pdf", sep = "")
      fiti <-
        mixture.penalized(Y,
                          lambda[i],
                          ini = ini,
                          method = method,
                          control = control)
      Gamma.path[i, , ] <- fiti$Gamma
      Alpha.path[i, ] <- fiti$Alpha
      Mu.path[i, , 1] <- fiti$Mu[, 2]
      Mu.path[i, , 2] <- fiti$Mu[, 3]
      Sigma1.path[i, , ] <- fiti$Sigma[, , 1]
      Sigma2.path[i, ] <- diag(fiti$Sigma[, , 2])
      Sigma3.path[i, ] <- diag(fiti$Sigma[, , 3])
      Pi.path[i, ] <- fiti$Pi
      loglike[i] <- fiti$loglike[length(fiti$loglike)]
      obj[i] <- fiti$obj[length(fiti$obj)]
      aic[i] <- fiti$aic
      bic[i] <- fiti$bic
      nout[i] <- fiti$nout
      df[i] <- fiti$df
      if (control$warm.start == TRUE) {
        ini <-
          list(
            Pi = fiti$Pi,
            Mu = fiti$Mu,
            Sigma = fiti$Sigma,
            Alpha = fiti$Alpha,
            Gamma = fiti$Gamma
          )
      }
      if (traceind == TRUE) {
        cat(
          "i = ",
          i,
          " lambda = ",
          round(lambda[i], 4),
          " AIC = ",
          round(aic[i], 4),
          " BIC = ",
          round(bic[i], 4),
          " nout = ",
          nout[i],
          "\n"
        )
      }
    }
    
    rval <-
      list(
        Gamma.path = Gamma.path,
        Alpha.path = Alpha.path,
        Mu.path = Mu.path,
        Sigma1.path = Sigma1.path,
        Sigma2.path = Sigma2.path,
        Sigma3.path = Sigma3.path,
        Pi.path = Pi.path,
        loglike.path = loglike,
        obj.path = obj,
        df.path = df,
        aic.path = aic,
        bic.path = bic,
        nout.path = nout
      )
    
    class(rval) <- "mixture.path"
    invisible(rval)
  }
